;;; vporshe1stcls.lisp --- Using 1st class generic function objects.

;; Author:        Didier Verna <didier@lrde.epita.fr>
;; Maintainer:    Didier Verna <didier@lrde.epita.fr>
;; Created:       Mon Apr  6 17:50:01 2009
;; Last Revision: Tue Apr  7 11:05:02 2009


;;; Commentary:

;; Contents management by FCM version 0.1.


;;; Code:


;; No more need for a PORSHE-VISITOR class, and neither for a
;; VISITABLE-PORSHE-COMPONENT (see comment below). So we're back to our
;; original hierarchy.

(defclass wheel ()
  ())

(defclass body ()
  ((wheels :initform (loop :for i :from 1 :upto 4
			   :for wheel = (make-instance 'wheel)
			   :collect wheel)
	   :accessor wheels)))

(defclass engine ()
  ())

(defclass porshe ()
  ((engine :initform (make-instance 'engine)
	   :accessor engine)
   (body :initform (make-instance 'body)
	 :accessor body)))


;; Explanation: in the following code:
;; (accept porshe (make-instance 'paint-car-visitor)),
;; we were creating a PAINT-CAR-VISITOR object just to propagate it to the
;; VISIT generic function, so that it can select the proper method. This is
;; one indirection for nothing. Indeed, since functions (even generic ones)
;; are first-class objects, we can let a function visit an object directly.

;; So, we don't need a VISIT generic function anymore, and the ACCEPT generic
;; function will recieve a generic function object directly as its second
;; argument.
(defgeneric accept (object visitor-function)
  ;; Note that for simple (non structural) objects, the ACCEPT function will
  ;; just call the VISITOR-FUNCTION on OBJECT, like this:
  ;; (defmethod accept ((wheel wheel) visitor-function)
  ;;   (funcall visitor-function wheel))
  ;; (defmethod accept ((engine engine) visitor-function)
  ;;   (funcall visitor-function engine))
  ;; This can be unified in a non-specialized method, default method.
  (:method (object visitor-function)
    (funcall visitor-function object))
  (:method ((body body) visitor-function)
    (funcall visitor-function body)
    (dolist (wheel (wheels body))
      (accept wheel visitor-function)))
  (:method ((porshe porshe) visitor-function)
    (funcall visitor-function porshe)
    (accept (engine porshe) visitor-function)
    (accept (body porshe) visitor-function)))

;; No more need for PAINT-PORSHE-VISITOR and RECYCLE-PORSHE-VISITOR. Instead,
;; we just need to define two generic (visiting) functions. The failure to
;; implement one of the methods below would trigger a run-time error.
(defgeneric paint (object)
  (:method ((wheel wheel))
    (format t "Painting a wheel.~%"))
  (:method ((body body))
    (format t "Painting the body.~%"))
  (:method ((engine engine))
    (format t "Painting the engine.~%"))
  (:method ((porshe porshe))
    (format t "Painting the Porshe.~%")))

(defgeneric recycle (object)
  (:method ((wheel wheel))
    (format t "Recycling a wheel.~%"))
  (:method ((body body))
    (format t "Recycling the body.~%"))
  (:method ((engine engine))
    (format t "Recycling the engine.~%"))
  (:method ((porshe porshe))
    (format t "Recycling the Porshe.~%")))


(defun main ()
  (let ((porshe (make-instance 'porshe)))
    (accept porshe #'paint)
    (accept porshe #'recycle))
  (quit))

;; And note that ACCEPT should just be named VISIT ;-)


;;; vporshe1stcls.lisp ends here
